/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "appl_locl.h"
#include <sl.h>

RCSID("$Id: vos.c,v 1.30 1998/03/14 23:55:30 assar Exp $");

static int empty_cmd (int, char **);
static int exa_cmd (int, char **);
static int help_cmd (int, char **);
static int quit_cmd (int, char **);
static int listp_cmd (int, char **);
static int livol_cmd (int, char **);
static int parti_cmd (int, char **);

static SL_cmd cmds[] = {
    {"addsite",    empty_cmd,       "not yet implemented"},
    {"apropos",    empty_cmd,       "not yet implemented"},
    {"backup",     empty_cmd,       "not yet implemented"},
    {"backupsys",  empty_cmd,       "not yet implemented"},
    {"changeaddr", empty_cmd,       "not yet implemented"},
    {"create",     empty_cmd,       "not yet implemented"},
    {"delentry",   empty_cmd,       "not yet implemented"},
    {"dump",       empty_cmd,       "not yet implemented"},
    {"examine",    exa_cmd,         "print information about a volume"},
    {"volinfo"},
    {"help",       help_cmd,        "print help"},
    {"?"},
    {"listpart",   listp_cmd,       "list partitions on a server"},
    {"listvldb",   empty_cmd,       "not yet implemented"},
    {"listvol",    livol_cmd,       "list volumes on a server"},
    {"lock",       empty_cmd,       "not yet implemented"},
    {"move",       empty_cmd,       "not yet implemented"},
    {"partinfo",   parti_cmd,       "print partition information on a server"},
    {"release",    empty_cmd,       "not yet implemented"},
    {"remove",     empty_cmd,       "not yet implemented"},
    {"remsite",    empty_cmd,       "not yet implemented"},
    {"rename",     empty_cmd,       "not yet implemented"},
    {"restore",    empty_cmd,       "not yet implemented"},
    {"status",     empty_cmd,       "not yet implemented"},
    {"syncserv",   empty_cmd,       "not yet implemented"},
    {"syncvldb",   empty_cmd,       "not yet implemented"},
    {"unlock",     empty_cmd,       "not yet implemented"},
    {"unlockvldb", empty_cmd,       "not yet implemented"},
    {"zap",        empty_cmd,       "not yet implemented"},
    {"quit",       quit_cmd,        "exit interactive mode"},
    {NULL}
};

/* this program needs __progname defined as a macro */
#define __progname "vos"
#define PROGNAME (vos_interactive ? "" : __progname" ")

/* if this is set the program runs in interactive mode */
static int vos_interactive = 0;


int printlistparts(char *server);
int printvolstat(char *volname,
		 const char *cell, const char *host, int noauth);
int printlistvol(char *server, int part);
int printpartinfo(char *server, char *partition);

static int
empty_cmd(int argc, char **argv)
{
    printf("%s%s has not been implemented yet!\n", PROGNAME, argv[0]);
    return 0;
}

static int
quit_cmd(int argc, char **argv)
{
    printf("exiting\n");
    return 1;
}

static int
help_cmd(int argc, char **argv)
{
    sl_help(cmds, argc, argv);
    return 0;
}


void
exa_usage()
{
    printf("Usage: %sexamine <volume name>\n", PROGNAME);    
}

static int
exa_cmd(int argc, char **argv)
{
    char *volname;
    char *cell = NULL;
    char *host = NULL;
    int noauth = 0 ;

    /* get rid of commandname */
    argc--;
    argv++;

    if (argc < 1) {
	exa_usage();
	return 0;
    }

    /* get rid of -id that some old people can write in front of volname */
    if (strcmp(argv[0], "-id") == 0) {
	argc--;
	argv++;
    }

    volname = argv[0];

    /* don't allow any bogus volname (assumes that argv[argc] == NULL) */
    if (volname == NULL || volname[0] == '-' || volname[0] == '\0') {
	exa_usage();
	return 0;
    }

    argv++;
    argc--;

    while(argc) {
	if (strcmp(argv[0], "-host") == 0 && argc > 0) {
	    host = argv[1] ;
	    argv++;
	    argc--;
	} else if (strcmp(argv[0], "-cell") == 0 && argc > 0) {
	    cell = argv[1] ;
	    argv++;
	    argc--;
	} else if (strcmp(argv[0], "-noauth") == 0) {
	    noauth = 1 ;
	} else {
	    exa_usage();
	    return 0 ;
	}
	argc--;
	argv++;
    }


    printvolstat(volname, cell, host, noauth);
    return 0;
}

void
listp_usage()
{
    printf("Usage: %slistpart <server name>\n", PROGNAME);
}

static int
listp_cmd(int argc, char **argv)
{
    char *server;

    /* get rid of command name */
    argc--;
    argv++;

    if (argc < 1) {
	listp_usage();
	return 0;
    }

    /* get rid of -server that some people may write */
    if (strcmp(argv[0], "-server") == 0) {
	argc--;
	argv++;
    }

    server = argv[0];

    /* sanity check (assumes that argv[argc] == NULL) */
    if (server == NULL || server[0] == '-' || server[0] == '\0') {
	listp_usage();
	return 0;
    }

    printlistparts(server);
    return 0;
}

void
livol_usage()
{
    printf("Usage: %slistvol <server name>\n", PROGNAME);
}

static int
livol_cmd(int argc, char **argv)
{
    char *server;

    /* remove command name */
    argc--;
    argv++;

    if (argc < 1) {
	livol_usage();
	return 0;
    }

    /* dinosaur protection */
    if (strcmp(argv[0], "-server") == 0) {
	argc--;
	argv++;
    }

    if (argc < 1) {
	livol_usage();
	return 0;
    }

    server = argv[0];

    /* sanity check (assumes that argv[argc] == NULL) */
    if (server == NULL || server[0] == '-' || server[0] == '\0') {
	livol_usage();
	return 0;
    }

    printlistvol(server, -1 ); /* XXX fix -1 */
    return 0;
}

void
parti_usage()
{
    printf("Usage: %spartinfo <server name> [partition name]\n", PROGNAME);
}

static int
parti_cmd(int argc, char **argv)
{
    char *server;
    char *partition = NULL;

    /* remove command name */
    argc--;
    argv++;

    if (argc < 1) {
	parti_usage();
	return 0;
    }

    /* dinosaur protection */
    if (strcmp(argv[0], "-server") == 0) {
	argc--;
	argv++;
    }

    server = argv[0];

    argc--;
    argv++;

    /* sanity check (assumes that argv[argc] == NULL)*/
    if (server == NULL || server[0] == '-' || server[0] == '\0') {
	parti_usage();
	return 0;
    }
    
    /* there could be a partition specified */
    if (argc > 0) {
	/* dinosaur protection */
	if (strcmp(argv[0], "-partition") == 0) {
	    argc--;
	    argv++;
	    partition = argv[0];
	    argc--;
	    argv++;
	} else
	    /* this could be the partition */
	    if (argv[0][0] != '-') {
		partition = argv[0];
		argc--;
		argv++;
	    }
    }

    printpartinfo(server, partition);
    return 0;
}

static char*
getvolumetype(int32_t flag)
{
    char *str ;
    if (flag & VLSF_RWVOL)
	str = "RW";
    else if (flag & VLSF_ROVOL)
	str = "RO";
    else if (flag & VLSF_BACKVOL)
	str = "BACKUP";
    else
	str = "FOO!";

    return str;
}

static char*
getvolumetype2(int32_t type)
{
    char *str = NULL;
    switch (type) {
    case RWVOL:
	str = "RW";
	break;
    case ROVOL:
	str = "RO";
	break;
    case BACKVOL:
	str = "BK";
	break;
    default:
	str = "FOO!";
    }
    return str;
}

int
printvolstat(char *volname, const char *cell, const char *host, int noauth)
{
    struct rx_connection *connvolser = NULL;
    struct rx_connection *connvldb = NULL;
    int error;
    int i;
    char *servername;
    char timestr[30];
    volEntries volint ;
    vldbentry vlentry;

    if (cell == NULL) 
	cell = cell_getthiscell();
    if (host == NULL)
	host = cell_findnamedbbyname (cell);

    
    connvldb = arlalib_getconnbyname(host,
				     7003,
				     VLDB_SERVICE_ID,
				     noauth);
    if (connvldb == NULL)
	return -1;
    
    
    error = VL_GetEntryByName(connvldb, volname, &vlentry);
    if (error) {
	fprintf(stderr, "VLDB: no such entry\n");
	return -1;
    } 

    connvolser = arlalib_getconnbyaddr(htonl(vlentry.serverNumber[0]),
				       NULL,
				       VOLSERVICE_PORT,
				       VOLSERVICE_ID,
				       noauth);
    if (connvolser == NULL)
	return -1 ;
    
    volint.val = NULL;
    if ((error = VOLSER_AFSVolListOneVolume(connvolser,
					    vlentry.serverPartition[0],
					    vlentry.volumeId[0],
					    &volint)) != 0) {
	printf("ListOneVolume failed with: %s (%d)\n", 
	       koerr_gettext(error),
	       error);
	return -1;
    }
    
    printf("%s\t\t\t%10u %s %8d K  %s\n", 
	   vlentry.name, 
	   vlentry.volumeId[0], 
	   getvolumetype(vlentry.serverFlags[0]),
	   volint.val->size,
	   volint.val->status == VOK ? "On-line" : "Busy");
    
    if (volint.val->status == VOK) {
	
	arlalib_getservername(htonl(vlentry.serverNumber[0]), &servername);
	printf("    %s /vicep%c\n", servername, 
	       'a' + vlentry.serverPartition[0]);
	printf("    ");
	free(servername);
	
	if (vlentry.flags & VLF_RWEXISTS)
	    printf("RWrite %u\t", vlentry.volumeId[RWVOL]);
    
	if (vlentry.flags & VLF_ROEXISTS )
	    printf("ROnly %u\t", vlentry.volumeId[ROVOL]);

	if (vlentry.flags & VLF_BACKEXISTS)
	    printf("Backup %u\t", vlentry.volumeId[BACKVOL]);

	printf("\n    MaxQuota %10d K\n", volint.val->maxquota);
	
	/* Get and format date */
	strncpy(timestr, 
		ctime( (time_t*) &volint.val->creationDate),
		sizeof(timestr));
	timestr[strlen(timestr)-1] = '\0';
	printf("    Creation    %s\n", timestr);
	
	/* Get and format date */
	strncpy(timestr, 
	       ctime((time_t *) &volint.val->updateDate),
	       sizeof(timestr));
	timestr[strlen(timestr)-1] = '\0';
	printf("    Last Update %s\n", timestr);

	printf("    %d accesses in the past day (i.e., vnode references)\n\n",
	       volint.val->dayUse);
	
    }

    /* Print volume stats */
    printf("    ");
    printf("RWrite: %u\t", vlentry.flags & VLF_RWEXISTS ? vlentry.volumeId[RWVOL] : 0);
    printf("ROnly: %u\t", vlentry.flags & VLF_ROEXISTS ? vlentry.volumeId[ROVOL] : 0);
    printf("Backup: %u\t", vlentry.flags & VLF_BACKEXISTS ? vlentry.volumeId[BACKVOL] : 0);

    printf("\n    number of sites -> %d\n", vlentry.nServers );
    
     for ( i = 0 ; i < vlentry.nServers ; i++) {
	 printf("       ");
	 
	 if (vlentry.serverNumber[i] != 0)
	     arlalib_getservername(htonl(vlentry.serverNumber[i]), &servername);
	 else
	     servername = strdup("error");
	 
	 printf("server %s partition /vicep%c %s Site\n",
		servername,
		'a' + vlentry.serverPartition[i],
		getvolumetype(vlentry.serverFlags[i]));
	 free(servername);
	 
     }
     
     free(volint.val);
     arlalib_destroyconn(connvolser);
     arlalib_destroyconn(connvldb);
     return 0;
}


int
getlistparts(char *host, struct pIDs *partIDs)
{
    struct rx_connection *connvolser = NULL;
    int error ;
    
    connvolser = arlalib_getconnbyname(host,
				       VOLSERVICE_PORT,
				       VOLSERVICE_ID,
				       0); /* XXX this means try auth */
    if (connvolser == NULL)
	return -1 ;
    
    if ((error = VOLSER_AFSVolListPartitions(connvolser,
					     partIDs)) != 0) {
	printf("getlistparts: ListPartitions failed with: %s (%d)\n", 
	       koerr_gettext(error),
	       error);
	return -1;
    }

    arlalib_destroyconn(connvolser);
    return 0;
}

int
printlistparts(char *host)
{
    struct pIDs partIDs;
    int i,j ;
    
    if ((i = getlistparts(host, &partIDs)) != 0)
	return i;

    j = 0 ;
    
    printf("The partitions on the server are:\n ");
    for (i = 0 ; i < 26 ; i++) {
	if (partIDs.partIds[i] != -1) {
	    j++;
	    printf("   /vicep%c%c", i + 'a', j % 6 == 0 ? '\n':' ');
	}
    }
    
    printf("\nTotal: %d\n", j);
    return 0;
}


int
printlistvol(char *host, int part)
{
    struct rx_connection *connvolser = NULL;
    volEntries volint ;
    struct pIDs partIDs;
    volintInfo *vi;
    int32_t partnum = 0;
    int32_t vol;
    int online, busy, offline;
    int error ;
    
    connvolser = arlalib_getconnbyname(host,
				       VOLSERVICE_PORT,
				       VOLSERVICE_ID,
				       0); /* XXX this means try auth */
    if (connvolser == NULL)
	return -1 ;

    if (part == -1) {
	if ((error = getlistparts(host, &partIDs)) != 0)
	    return -1;
    } else {
	for (partnum = 0 ; partnum < 26 ; partnum++)
	    partIDs.partIds[partnum] = -1 ;
	partIDs.partIds[0] = part ;
    }

    partnum = 0;
    while(1) {
	while(partIDs.partIds[partnum] == -1 && partnum < 26 ) partnum++;
	if (partnum >= 26) 
	    break;
	
	volint.val = NULL;
	if ((error = VOLSER_AFSVolListVolumes(connvolser,
					      partIDs.partIds[partnum],
					      1, /* We want full info */
					      &volint)) != 0) {
	    printf("printlistvol: PartitionInfo failed with: %s (%d)\n", 
		   koerr_gettext(error),
		   error);
	    return -1;
	}
	online =  busy = offline = 0;

	printf("Total number of volumes on server %s partition /vicep%c: %d\n",
	       host, 
	       partnum + 'a',
	       volint.len);

	for (vol = 0 ; vol < volint.len ; vol ++) {
	    vi = &volint.val[vol];
	    
	    if (vi->status == VOK)
		online++;
	    else
		offline++;

	    printf("%-38s %10u %s %10u K %s\n", 
		   vi->name,
		   vi->volid,
		   getvolumetype2(vi->type),
		   vi->size,
		   vi->status == VOK ? "On-line" : "Off-line");

	}

	printf("\nTotal volumes onLine %d ; Total volumes offLine %d " \
	       "; Total busy %d\n\n",
	       online, busy,offline);
	      
	free(volint.val);
	partnum++;
    }
    
    arlalib_destroyconn(connvolser);
    return 0;
}

int
printpartinfo(char *host, char *part)
{
    struct rx_connection *connvolser = NULL;
    struct diskPartition partinfo ;
    struct pIDs partIDs;
    int iter = 0;
    int partnum = 0;
    char name[30];
    int error ;
    
    connvolser = arlalib_getconnbyname(host,
				       VOLSERVICE_PORT,
				       VOLSERVICE_ID,
				       0); /* XXX This means try auth */
    if (connvolser == NULL)
	return -1 ;

    if (part == NULL) {
	iter = 1;
	if ((error = getlistparts(host, &partIDs)) != 0)
	    return error;
    } else {
	if (strlen(part) <= 2) {
	    snprintf(name, sizeof(name), "/vicep%s", part);
	    part = name;
	}
    }

    while(part != NULL || iter) {
    
	if (iter) {
	    while(partIDs.partIds[partnum] == -1 && partnum < 26 ) partnum++;
	    if (partnum < 26) {
		snprintf(name, sizeof(name), "/vicep%c", 'a' + partnum);
		part = name;
		partnum++;
	    } else {
		iter = 0 ;
		continue;
	    }
	}   

	if ((error = VOLSER_AFSVolPartitionInfo(connvolser,
						part,
						&partinfo)) != 0) {
	    printf("printpartinfo: PartitionInfo failed with: %s (%d)\n", 
		   koerr_gettext(error),
		   error);
	    return -1;
	}
	printf("Free space on partition %s %d K blocks out of total %d\n",
	       partinfo.name, 
	       partinfo.free,
	       partinfo.totalUsable);
	part = NULL;
    }
    
    arlalib_destroyconn(connvolser);
    return 0;
}

int
main(int argc, char **argv)
{
    int ret = 0;
    
    cell_init(0);

    if (argc > 1)
	ret = sl_command(cmds, argc - 1, argv + 1);
    else {
	vos_interactive = 1;
	printf("vos - an arla tool for administrating AFS volumes.\n");
	printf("Type \"help\" to get a list of commands.\n");
	ret = sl_loop(cmds, __progname": ");  
    }
    return ret;
}
